@@ -2,29 +2,36 @@
 # pylint: disable=missing-module-docstring
 from __future__ import annotations
 
-import io
 import logging
+import re
 import subprocess
 import sys
-from collections import defaultdict, deque
-from configparser import RawConfigParser
-from pathlib import Path
+from contextlib import suppress
 from copy import deepcopy
-from packaging.version import Version
+from pathlib import Path
+from shutil import copyfile, copytree, rmtree
 
 import asyncclick as click
 import git
 import tomlkit
-from moat.util import P, add_repr, attrdict, make_proc, yload, yprint
-from packaging.requirements import Requirement
 from attrs import define, field
-from shutil import rmtree, copyfile, copytree
-from contextlib import suppress
+from packaging.requirements import Requirement
+from packaging.version import Version
+
+from moat.util import P, add_repr, attrdict, make_proc, yload, yprint
+from moat.util.exec import run as run_
+
+from collections import defaultdict, deque
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from collections.abc import Sequence
 
 logger = logging.getLogger(__name__)
 
 PACK = Path("packaging")
-ARCH = subprocess.check_output(["dpkg", "--print-architecture"]).decode("utf-8").strip()
+ARCH = subprocess.check_output(["/usr/bin/dpkg", "--print-architecture"]).decode("utf-8").strip()
+SRC = re.compile(r"^Source:\s+(\S+)\s*$", re.MULTILINE)
 
 
 def dash(n: str) -> str:
@@ -44,14 +51,14 @@ def undash(n: str) -> str:
 
 
 class ChangedError(RuntimeError):
-    def __init__(subsys, tag, head):
+    def __init__(self, subsys, tag, head):
         self.subsys = subsys
         self.tag = tag
         self.head = head
 
     def __str__(self):
         s = self.subsys or "Something"
-        if head is None:
+        if (head := self.head) is None:
             head = "HEAD"
         else:
             head = head.hexsha[:9]
@@ -105,7 +112,7 @@ class Package(_Common):
 
     @property
     def verstr(self):
-        v=self.vers
+        v = self.vers
         return f"{v.tag}-{v.pkg}"
 
     @property
@@ -143,18 +150,12 @@ class Package(_Common):
     def mdash(self):
         return dash(self.name)
 
-    def populate(self, path: Path, real=None):
-        """
-        Collect this package's file names.
-        """
-        self.path = path
-        for fn in path.iterdir():
-            if fn.name == "__pycache__":
-                continue
-            if (sb := self.subs.get(fn.name, None)) is not None:
-                sb.populate(fn, real=self if sb.hidden else None)
-            else:
-                (real or self).files.add(fn)
+    @property
+    def srcname(self):
+        ctl = PACK / self.dash / "debian" / "control"
+        src = ctl.read_text()
+        sm = SRC.match(src)
+        return sm.group(1)
 
     def copy(self) -> None:
         """
@@ -186,15 +187,13 @@ class Package(_Common):
         between the head and the @tag commit
         """
         head = self._repo.head.commit
-        try:
-            lc = self.last_commit
-        except AttributeError:
+        if not hasattr(self, "last_commit"):
             return True
         for d in head.diff(
-                self.last_commit if main else self._repo.last_tag,
-                paths=self.path if main else Path("packaging")/self.dash,
-            ):
-            pp=Path(d.b_path)
+            self.last_commit if main else self._repo.last_tag,
+            paths=self.path if main else Path("packaging") / self.dash,
+        ):
+            pp = Path(d.b_path)
             if pp.name == "changelog" and pp.parent.name == "debian":
                 continue
             if (
@@ -212,9 +211,9 @@ class Repo(git.Repo, _Common):
     moat_tag = None
     _last_tag = None
 
-    toplevel:str
+    toplevel: str
 
-    def __init__(self, toplevel:str, *a, **k):
+    def __init__(self, toplevel: str, *a, **k):
         self.toplevel = toplevel
 
         super().__init__(*a, **k)
@@ -234,7 +233,6 @@ class Repo(git.Repo, _Common):
             self.versions = yload(f, attr=True)
         self.orig_versions = deepcopy(self.versions)
 
-
     def write_tags(self):
         if self.versions == self.orig_versions:
             logger.warning("No changed versions. Not tagging.")
@@ -245,7 +243,7 @@ class Repo(git.Repo, _Common):
         return True
 
     @property
-    def last_tag(self) -> Tag | None:
+    def last_tag(self) -> git.Tag | None:
         """
         Return the most-recent tag for this repo
         """
@@ -285,7 +283,7 @@ class Repo(git.Repo, _Common):
     def parts(self):
         return self._repos.values()
 
-    def tags_of(self, c: Commit) -> Sequence[Tag]:
+    def tags_of(self, c: git.Commit) -> Sequence[git.Tag]:
         return self._commit_tags[c]
 
     def _add_repo(self, name):
@@ -309,7 +307,24 @@ class Repo(git.Repo, _Common):
                 continue
             self._add_repo(str(fn.name))
 
-        self._repos[self.toplevel].populate(Path(self.toplevel))
+        res = subprocess.run(
+            ["/usr/bin/git", "ls-files", "-z", "--exclude-standard"],
+            check=False,
+            stdout=subprocess.PIPE,
+        )
+        for fn in res.stdout.split(b"\0"):
+            if not fn:
+                continue
+            fn = Path(fn.decode("utf-8"))  # noqa:PLW2901
+            if fn.name == ".gitignore":  # heh
+                continue
+            sb = self.repo_for(fn, True)
+            if sb is None:
+                continue
+            sb = dash(sb)
+            if sb not in self._repos:
+                raise RuntimeError(f"Inconsistent repo data: {sb} not found")
+            self._repos[sb].files.add(fn)
 
     def repo_for(self, path: Path | str, main: bool | None) -> str:
         """
@@ -326,12 +341,15 @@ class Repo(git.Repo, _Common):
 
         name = path.parts[0]
         if main is not False and name == self.toplevel:
+            res = name
             for p in path.parts[1:]:
                 if p not in sc.subs:
                     break
-                name += "." + p
                 sc = sc.subs[p]
-            return name
+                name += "." + p
+                if not sc.hidden:
+                    res = name
+            return res
 
         return None
 
@@ -361,17 +379,19 @@ class Repo(git.Repo, _Common):
         Test whether any subsystem changed since the "tagged" commit
 
         """
-        if tag is None:
-            tag = self.last_tag
+        tag = self.last_tag
         head = self._repo.head.commit
-        print("StartDiff B",self,tag,head,file=sys.stderr)
+        print("StartDiff B", self, tag, head, file=sys.stderr)
         for d in head.diff(tag):
-            if self.repo_for(d.a_path, main) == self.toplevel and self.repo_for(d.b_path, main) == self.toplevel:
+            if (
+                self.repo_for(d.a_path, main) == self.toplevel
+                and self.repo_for(d.b_path, main) == self.toplevel
+            ):
                 continue
             return True
         return False
 
-    def tagged(self, c: Commit = None) -> Tag | None:
+    def tagged(self, c: git.Commit = None) -> git.Tag | None:
         """Return a commit's tag name.
         Defaults to the head commit.
         Returns None if no tag, raises ValueError if more than one is found.
@@ -387,8 +407,6 @@ class Repo(git.Repo, _Common):
         if not tt:
             return None
         if len(tt) > 1:
-            if subsys is not None:
-                raise ValueError(f"Multiple tags for {subsys}: {tt}")
             raise ValueError(f"Multiple tags: {tt}")
         return tt[0].name
 
@@ -401,13 +419,38 @@ async def cli():
     pass
 
 
+@cli.command("rerepo")
+@click.option("-a", "--all", is_flag=True, help="Move all your repositories.")
+@click.argument("names", type=str, nargs=-1)
+@click.pass_obj
+async def move_repo(obj, **kw):
+    """Move from forge A to forge B.
+
+    This command moves your repositories from A (github …) to a local copy,
+    B (codeberg,or any other forgejo instance), and/or Radicle.
+
+    It adds a "migrated" branch and (optionally) deletes
+    all other branches and tags.
+
+    By default, the '--all' option does not touch forked repositores.
+
+    For access tokens and configuration, load a private config file.
+    See `moat util cfg -l moat.src src` for defaults.
+
+    If the local copy is present, it will be refreshed via `git fetch`.
+    """
+    from .move import mv_repos  # noqa: PLC0415
+
+    await mv_repos(obj.cfg.src.move, **kw)
+
+
 def fix_deps(deps: list[str], tags: dict[str, str]) -> bool:
     """Adjust dependencies"""
     work = False
     for i, dep in enumerate(deps):
         r = Requirement(dep)
         if r.name in tags:
-            dep = f"{r.name} ~= {tags[r.name]}"
+            dep = f"{r.name} ~= {tags[r.name]}"  # noqa:PLW2901
             if deps[i] != dep:
                 deps[i] = dep
                 work = True
@@ -428,12 +471,12 @@ def run_tests(pkg: str | None, *opts) -> bool:
         return True
     try:
         print("\n*** Testing:", pkg)
-        subprocess.run(
-            ["python3", "-mpytest", *opts, tests],
+        subprocess.run(  # noqa:S603
+            ["/usr/bin/python3", "-mpytest", *opts, tests],
+            check=False,
             stdin=sys.stdin,
             stdout=sys.stdout,
             stderr=sys.stderr,
-            check=True,
         )
     except subprocess.CalledProcessError:
         return False
@@ -507,18 +550,18 @@ def default_dict(a, b, c, cls=dict, repl=lambda x: x) -> dict:
                 mod = True
             if va:
                 for vv in va:
-                    vv = repl(vv)
+                    vv = repl(vv)  # noqa:PLW2901
                     if vv not in vb:
                         vb.insert(0, vv)
                         mod = True
             if vc:
                 for vv in vc:
-                    vv = repl(vv)
+                    vv = repl(vv)  # noqa:PLW2901
                     if vv not in vb:
                         vb.insert(0, vv)
                         mod = True
         else:
-            v = repl(va) or vb or repl(vc)
+            v = repl(va) or vb or repl(vc)  # noqa:PLW2901
             if vb != v:
                 b[k] = v
                 mod = True
@@ -715,7 +758,7 @@ def tags(obj):
     for r in repo.parts:
         try:
             tag = r.last_tag
-        except KeyError:
+        except AttributeError:
             continue
         if r.has_changes(True):
             print(f"{r.dash} {tag} STALE")
@@ -870,7 +913,7 @@ async def build(
     if g_done is not None:
         g_done = Path(g_done)
     else:
-        g_done = Path("/tmp/nonexistent")
+        g_done = Path("/tmp/nonexistent")  # noqa:S108
     repo = Repo(cfg.src.toplevel, None)
 
     tags = dict(version)
@@ -907,8 +950,8 @@ async def build(
     for name in PACK.iterdir():
         if name.suffix != ".changes":
             continue
-        name = name.stem
-        name, vers, _ = name.split("_")
+        name = name.stem  # noqa:PLW2901
+        name, vers, _ = name.split("_")  # noqa:PLW2901
         debversion[name] = vers.rsplit("-", 1)[0]
 
     # Step 0: basic check
@@ -933,13 +976,13 @@ async def build(
                     pkg=1,
                     rev=repo.head.commit.hexsha,
                 )
-                logger.debug("Changes: %s %s",r.name,r.verstr)
+                logger.debug("Changes: %s %s", r.name, r.verstr)
             elif r.has_changes(False):
                 r.vers.pkg += 1
                 r.vers.rev = repo.head.commit.hexsha
-                logger.debug("Build Changes: %s %s",r.name,r.verstr)
+                logger.debug("Build Changes: %s %s", r.name, r.verstr)
             else:
-                logger.debug("No Changes: %s %s",r.name,r.verstr)
+                logger.debug("No Changes: %s %s", r.name, r.verstr)
 
     elif not no_tag:
         err = set()
@@ -983,31 +1026,35 @@ async def build(
     for r in repos:
         rd = PACK / r.dash
         p = rd / "pyproject.toml"
-        if not p.is_file():
+        skip = rd / "SKIP.build"
+
+        if skip.is_file():
             # bad=True
             print("Skip:", r.name, file=sys.stderr)
             continue
-        with p.open("r") as f:
-            pr = tomlkit.load(f)
-            pr["project"]["version"] = r.last_tag
 
-        if not no_version:
-            try:
-                deps = pr["project"]["dependencies"]
-            except KeyError:
-                pass
-            else:
-                fix_deps(deps, tags)
-            try:
-                deps = pr["project"]["optional_dependencies"]
-            except KeyError:
-                pass
-            else:
-                for v in deps.values():
-                    fix_deps(v, tags)
+        if p.is_file():
+            with p.open("r") as f:
+                pr = tomlkit.load(f)
+                pr["project"]["version"] = r.last_tag
 
-        p.write_text(pr.as_string())
-        repo.index.add(p)
+            if not no_version:
+                try:
+                    deps = pr["project"]["dependencies"]
+                except KeyError:
+                    pass
+                else:
+                    fix_deps(deps, tags)
+                try:
+                    deps = pr["project"]["optional_dependencies"]
+                except KeyError:
+                    pass
+                else:
+                    for v in deps.values():
+                        fix_deps(v, tags)
+            p.write_text(pr.as_string())
+
+            repo.index.add(p)
 
     # Step 3: copy to packaging dir
     for r in repos:
@@ -1020,66 +1067,66 @@ async def build(
 
         for r in repos:
             ltag = r.last_tag
-            if r.vers.get("deb","-") == f"{ltag}-{r.vers.pkg}":
+            if r.vers.get("deb", "-") == f"{ltag}-{r.vers.pkg}":
                 continue
             rd = PACK / r.dash
             p = rd / "debian"
             if not p.is_dir():
                 continue
-            if not (rd/"debian"/"changelog").exists():
-                subprocess.run(
-                    [
-                        "debchange",
-                        "--create",
-                        "--newversion", f"{r.last_tag}-{r.vers.pkg}",
-                        "--package", r.mdash,
-                        f"Initial release for {forcetag}",
-                     ],
+            if not (rd / "debian" / "changelog").exists():
+                await run_(
+                    "debchange",
+                    "--create",
+                    "--newversion",
+                    f"{r.last_tag}-{r.vers.pkg}",
+                    "--package",
+                    r.mdash,
+                    f"Initial release for {forcetag}",
                     cwd=rd,
-                    check=True,
-                    stdout=sys.stdout,
-                    stderr=sys.stderr,
                 )
 
             try:
-                res = subprocess.run(
-                    ["dpkg-parsechangelog", "-l", "debian/changelog", "-S", "version"],
+                res = await run_(
+                    "dpkg-parsechangelog",
+                    "-l",
+                    "debian/changelog",
+                    "-S",
+                    "version",
                     cwd=rd,
-                    check=True,
-                    stdout=subprocess.PIPE,
+                    capture=True,
                 )
-                tag,ptag = res.stdout.strip().decode("utf-8").rsplit("-", 1)
+                tag, ptag = res.strip().rsplit("-", 1)
                 ptag = int(ptag)
                 if tag != ltag or r.vers.pkg > ptag:
-                    subprocess.run(
-                        [
-                            "debchange",
-                            "--distribution",
-                            "unstable",
-                            "--newversion",
-                            f"{ltag}-{r.vers.pkg}",
-                            f"New release for {forcetag}",
-                        ],
+                    await run_(
+                        "debchange",
+                        "--distribution",
+                        "unstable",
+                        "--newversion",
+                        f"{ltag}-{r.vers.pkg}",
+                        f"New release for {forcetag}",
                         cwd=rd,
-                        check=True,
                     )
                     repo.index.add(p / "changelog")
 
                 elif tag == ltag and r.vers.pkg < ptag:
                     r.vers.pkg = ptag
 
-                changes = PACK / f"{r.mdash}_{ltag}-{r.vers.pkg}_{ARCH}.changes"
-                if debversion.get(r.dash, "") != ltag or r.vers.pkg != ptag or test_chg and not changes.exists():
-
-                    subprocess.run(["debuild", "--build=binary"] + deb_opts, cwd=rd, check=True)
+                changes = PACK / f"{r.srcname}_{ltag}-{r.vers.pkg}_{ARCH}.changes"
+                if (
+                    debversion.get(r.dash, "") != ltag
+                    or r.vers.pkg != ptag
+                    or (test_chg and not changes.exists())
+                ):
+                    await run_("debuild", "--build=binary", *deb_opts, cwd=rd, echo=obj.debug > 1)
             except subprocess.CalledProcessError:
                 if not run:
                     print("*** Failure packaging", r.name, file=sys.stderr)
                 else:
                     print("Failure packaging", r.name, file=sys.stderr)
-                    no_commit=True
-                    no_deb=True
-                    no_pypi=True
+                    no_commit = True
+                    no_deb = True
+                    no_pypi = True
 
     # Step 5: build PyPI package
     if not no_pypi:
@@ -1092,7 +1139,7 @@ async def build(
                 continue
             tag = r.last_tag
             name = r.dash
-            if r.vers.get("pypi","-") == r.last_tag:
+            if r.vers.get("pypi", "-") == r.last_tag:
                 continue
 
             targz = rd / "dist" / f"{r.under}-{tag}.tar.gz"
@@ -1104,7 +1151,7 @@ async def build(
                 up.add(r)
             else:
                 try:
-                    subprocess.run(["python3", "-mbuild", "-snw"], cwd=rd, check=True)
+                    await run_("python3", "-mbuild", "-snw", cwd=rd, echo=obj.debug > 1)
                 except subprocess.CalledProcessError:
                     err.add(r.name)
                 else:
@@ -1117,8 +1164,8 @@ async def build(
                 print("Build errors:", file=sys.stderr)
                 print(*err, file=sys.stderr)
                 print("Please fix and try again.", file=sys.stderr)
-                no_commit=True
-                no_deb=True
+                no_commit = True
+                no_deb = True
 
         # Step 6: upload PyPI package
         elif run:
@@ -1133,10 +1180,13 @@ async def build(
                 targz = Path("dist") / f"{r.under}-{tag}.tar.gz"
                 whl = Path("dist") / f"{r.under}-{tag}-py3-none-any.whl"
                 try:
-                    res = subprocess.run(
-                        ["twine", "upload", str(targz), str(whl)],
+                    await run_(
+                        "twine",
+                        "upload",
+                        str(targz),
+                        str(whl),
                         cwd=rd,
-                        check=True,
+                        echo=obj.debug > 1,
                     )
                 except subprocess.CalledProcessError:
                     err.add(r.name)
@@ -1147,8 +1197,8 @@ async def build(
                 print("Upload errors:", file=sys.stderr)
                 print(*err, file=sys.stderr)
                 print("Please fix(?) and try again.", file=sys.stderr)
-                no_commit=True
-                no_deb=True
+                no_commit = True
+                no_deb = True
             else:
                 r.vers.pypi = r.last_tag
 
@@ -1159,12 +1209,12 @@ async def build(
             dput_opts = ["-u", "ext"]
         for r in repos:
             ltag = r.last_tag
-            if r.vers.get("deb","-") == f"{ltag}-{r.vers.pkg}":
+            if r.vers.get("deb", "-") == f"{ltag}-{r.vers.pkg}":
                 continue
             if not (PACK / r.dash / "debian").is_dir():
                 continue
-            changes = PACK / f"{r.mdash}_{ltag}-{r.vers.pkg}_{ARCH}.changes"
-            done = PACK / f"{r.mdash}_{ltag}-{r.vers.pkg}_{ARCH}.done"
+            changes = PACK / f"{r.srcname}_{ltag}-{r.vers.pkg}_{ARCH}.changes"
+            done = PACK / f"{r.srcname}_{ltag}-{r.vers.pkg}_{ARCH}.done"
             if done.exists():
                 continue
             if g_done is not None:
@@ -1172,7 +1222,7 @@ async def build(
                 if gdone.exists():
                     continue
             try:
-                subprocess.run(["dput", *dput_opts, str(changes)], check=True)
+                await run_("dput", *dput_opts, str(changes), echo=obj.debug > 1)
             except subprocess.CalledProcessError:
                 err.add(r.name)
             else:
@@ -1181,7 +1231,7 @@ async def build(
             print("Upload errors:", file=sys.stderr)
             print(*err, file=sys.stderr)
             print("Please fix(?) and try again.", file=sys.stderr)
-            no_commit=True
+            no_commit = True
         else:
             r.vers.deb = f"{ltag}-{r.vers.pkg}"
 
