@@ -27,15 +27,15 @@ class CancelTracer:
     def __init__(self):
         pass
 
-    def skip(self, scope):
-        if scope._stack is None:
+    def skip(self, scope):  # noqa: D102
+        if scope._stack is None:  # noqa: SLF001
             return True
-        if scope._stack[5].f_code.co_name == "connect_tcp":
+        if scope._stack[5].f_code.co_name == "connect_tcp":  # noqa: SLF001
             return True
         return False
 
-    def scope_entered(self, scope):
-        scope._stack = s = []
+    def scope_entered(self, scope):  # noqa: D102
+        scope._stack = s = []  # noqa: SLF001
         f = inspect.currentframe().f_back
         while f:
             s.append(f)
@@ -46,20 +46,20 @@ class CancelTracer:
 
         debug("EnterCS %r", scope)
 
-    def scope_exited(self, scope):
+    def scope_exited(self, scope):  # noqa: D102
         if self.skip(scope):
             return
         debug("ExitCS %r", scope)
 
-    def scope_cancelled(self, scope, reason):
+    def scope_cancelled(self, scope, reason):  # noqa: D102
         if self.skip(scope):
             return
         #       if reason.value == 0:
         #           breakpoint()
         debug("KillCS %r %s", scope, reason.name)
 
-    def task_spawned(self, task):
+    def task_spawned(self, task):  # noqa: D102
         debug("EnterT %r", task)
 
-    def task_exited(self, task):
+    def task_exited(self, task):  # noqa: D102
         debug("ExitT %r", task)
